@@ -22,6 +22,7 @@
 #include "jsquery.h"
 
 static ExtractedNode *recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path);
+static ExtractedNode *makeAnyNode(bool not, bool indirect, PathItem *path);
 static int coundChildren(ExtractedNode *node, ExtractedNodeType type, bool first, bool *found);
 static void fillChildren(ExtractedNode *node, ExtractedNodeType type, bool first, ExtractedNode **items, int *i);
 static void flatternTree(ExtractedNode *node);
@@ -48,18 +49,32 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 	ExtractedNodeType	type;
 	JsQueryItem			elem, e;
 
+	e.hint = false;
 	check_stack_depth();
 
 	switch(jsq->type)
 	{
 		case jqiAnd:
 		case jqiOr:
-			type = ((jsq->type == jqiAnd) == not) ? eOr : eAnd;
+		case jqiFilter:
+			type = ((jsq->type == jqiAnd || jsq->type == jqiFilter) == not) ? eOr : eAnd;
 
-			jsqGetLeftArg(jsq, &elem);
-			leftNode = recursiveExtract(&elem, not, false, path);
-			jsqGetRightArg(jsq, &elem);
-			rightNode = recursiveExtract(&elem, not, false, path);
+			if (jsq->type == jqiFilter)
+			{
+				jsqGetArg(jsq, &elem);
+				leftNode = recursiveExtract(&elem, not, false, path);
+				if (jsqGetNext(jsq, &elem))
+					rightNode = recursiveExtract(&elem, not, false, path);
+				else
+					rightNode = makeAnyNode(not, false, path);
+			}
+			else
+			{
+				jsqGetLeftArg(jsq, &elem);
+				leftNode = recursiveExtract(&elem, not, false, path);
+				jsqGetRightArg(jsq, &elem);
+				rightNode = recursiveExtract(&elem, not, false, path);
+			}
 
 			if (!leftNode || !rightNode)
 			{
@@ -94,7 +109,8 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 			pathItem->type = iKey;
 			pathItem->s = jsqGetString(jsq, &pathItem->len);
 			pathItem->parent = path;
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, pathItem);
 			return recursiveExtract(&elem, not, indirect, pathItem);
 		case jqiAny:
 		case jqiAll:
@@ -103,14 +119,16 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 			pathItem = (PathItem *)palloc(sizeof(PathItem));
 			pathItem->type = iAny;
 			pathItem->parent = path;
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, pathItem);
 			return recursiveExtract(&elem, not, true, pathItem);
 		case jqiIndexArray:
 			pathItem = (PathItem *)palloc(sizeof(PathItem));
 			pathItem->type = iIndexArray;
 			pathItem->arrayIndex = jsq->arrayIndex;
 			pathItem->parent = path;
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, pathItem);
 			return recursiveExtract(&elem, not, true, pathItem);
 		case jqiAnyArray:
 		case jqiAllArray:
@@ -119,7 +137,8 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 			pathItem = (PathItem *)palloc(sizeof(PathItem));
 			pathItem->type = iAnyArray;
 			pathItem->parent = path;
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, pathItem);
 			return recursiveExtract(&elem, not, true, pathItem);
 		case jqiAnyKey:
 		case jqiAllKey:
@@ -128,10 +147,12 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 			pathItem = (PathItem *)palloc(sizeof(PathItem));
 			pathItem->type = iAnyKey;
 			pathItem->parent = path;
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, pathItem);
 			return recursiveExtract(&elem, not, true, pathItem);
 		case jqiCurrent:
-			jsqGetNext(jsq, &elem);
+			if (!jsqGetNext(jsq, &elem))
+				return makeAnyNode(not, indirect, path);
 			return recursiveExtract(&elem, not, indirect, path);
 		case jqiEqual:
 			if (not)
@@ -229,16 +250,18 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 
 			if (jsq->type == jqiGreater || jsq->type == jqiGreaterOrEqual)
 			{
+				result->bounds.leftBound = (JsQueryItem *)palloc(sizeof(JsQueryItem));
 				result->bounds.leftInclusive = (jsq->type == jqiGreaterOrEqual);
 				result->bounds.rightBound = NULL;
-				result->bounds.leftBound = (JsQueryItem *)palloc(sizeof(JsQueryItem));
+				result->bounds.rightInclusive = false;
 				jsqGetArg(jsq, result->bounds.leftBound);
 			}
 			else
 			{
+				result->bounds.rightBound = (JsQueryItem *)palloc(sizeof(JsQueryItem));
 				result->bounds.rightInclusive = (jsq->type == jqiLessOrEqual);
 				result->bounds.leftBound = NULL;
-				result->bounds.rightBound = (JsQueryItem *)palloc(sizeof(JsQueryItem));
+				result->bounds.leftInclusive = false;
 				jsqGetArg(jsq, result->bounds.rightBound);
 			}
 			return result;
@@ -261,6 +284,25 @@ recursiveExtract(JsQueryItem *jsq, bool not, bool indirect, PathItem *path)
 	return NULL;
 }
 
+/*
+ * Make node for checking existence of path.
+ */
+static ExtractedNode *
+makeAnyNode(bool not, bool indirect, PathItem *path)
+{
+	ExtractedNode  *result;
+
+	if (not)
+		return NULL;
+
+	result = (ExtractedNode *) palloc(sizeof(ExtractedNode));
+	result->type = eAny;
+	result->hint = false;
+	result->path = path;
+	result->indirect = indirect;
+	return result;
+}
+
 /*
  * Count number of children connected with nodes of same type.
  */
@@ -659,12 +701,13 @@ makeEntries(ExtractedNode *node, MakeEntryHandler handler, Pointer extra)
 		for (i = 0; i < node->args.count; i++)
 		{
 			child = node->args.items[i];
-			if (!child) continue;
-			if (child->sClass > node->sClass && !child->forceIndex)
-			{
-				Assert(node->type != eOr);
+			if (!child)
+				continue;
+			/* Skip non-selective AND children */
+			if (child->sClass > node->sClass &&
+				node->type == eAnd &&
+				!child->forceIndex)
 				continue;
-			}
 			child = makeEntries(child, handler, extra);
 			if (child)
 			{
@@ -731,6 +774,8 @@ setSelectivityClass(ExtractedNode *node, CheckEntryHandler checkHandler,
 				if (!child)
 					continue;
 
+				setSelectivityClass(child, checkHandler, extra);
+
 				if (!isLogicalNodeType(child->type))
 				{
 					if (child->hint == jsqNoIndex ||
@@ -738,8 +783,6 @@ setSelectivityClass(ExtractedNode *node, CheckEntryHandler checkHandler,
 						continue;
 				}
 
-				setSelectivityClass(child, checkHandler, extra);
-
 				if (child->forceIndex)
 					node->forceIndex = true;
 
@@ -825,7 +868,7 @@ execRecursiveTristate(ExtractedNode *node, GinTernaryValue *check)
 			res = GIN_TRUE;
 			for (i = 0; i < node->args.count; i++)
 			{
-				v = execRecursive(node->args.items[i], check);
+				v = execRecursive(node->args.items[i], (bool *) check);
 				if (v == GIN_FALSE)
 					return GIN_FALSE;
 				else if (v == GIN_MAYBE)
@@ -836,7 +879,7 @@ execRecursiveTristate(ExtractedNode *node, GinTernaryValue *check)
 			res = GIN_FALSE;
 			for (i = 0; i < node->args.count; i++)
 			{
-				v = execRecursive(node->args.items[i], check);
+				v = execRecursive(node->args.items[i], (bool *) check);
 				if (v == GIN_TRUE)
 					return GIN_TRUE;
 				else if (v == GIN_MAYBE)
@@ -854,6 +897,11 @@ execRecursiveTristate(ExtractedNode *node, GinTernaryValue *check)
 static void
 debugPath(StringInfo buf, PathItem *path)
 {
+	if (!path)
+	{
+		appendStringInfoChar(buf, '$');
+		return;
+	}
 	if (path->parent)
 	{
 		debugPath(buf, path->parent);
